/* 2003/2004 Erik Andresen
 * erik@defiant.homedns.org
 *
 * The serial part is copied from miniterm (see the big comment below)
 * Thanks to the Serial Programming HOWTO
 * http://www.tldp.org/HOWTO/Serial-Programming-HOWTO/
 * 
 * You can read more about intelligent interface programming here:
 * http://www.fischertechnik.de/computing/pdf/schnittstelle.pdf
 */


/*
 *  AUTHOR: Sven Goldt (goldt@math.tu-berlin.de)
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version 2
 *  of the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
*/
/*
 This is like all programs in the Linux Programmer's Guide meant
 as a simple practical demonstration.
 It can be used as a base for a real terminal program.
*/


#include <termios.h>
#include <stdio.h>
#include <unistd.h>
#include <fcntl.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <pthread.h>

#define BAUDRATE B9600
#define TRUE 1
#define FALSE 0

/* set this Value higher if you loose packets, eg 300000 is a better value, but slower */
#define INTERFACE_QUERY_TIME 10000  /* waiting 250000 microseconds = 0,25 seconds */

#define _POSIX_SOURCE 1 /* POSIX compliant source */

typedef struct IInterface {
	int IInt; /* The variable that points to the interface */
	char numIfaces;
	struct termios newioset; /* Our new settings */
	struct termios saveioset; /* The pointer we save the old settings in*/
	unsigned char MotorStatus; /* hex, Motor state */
	unsigned short int Digitals; /* all Digitals in a hex number */
	unsigned char MotorStatus2; /* extension module */
	unsigned short int Digitals2;
	unsigned short int AnaX; /* Analog X, dezimal */
	unsigned short int AnaY;
	char Digital[16+1]; /* array for digital inputs */
} IInterfaceDef;


/* used before exiting... -just clean up */
void IIexit(IInterfaceDef **myInterface)
{
	tcsetattr((*myInterface)->IInt, TCSANOW, &(*myInterface)->saveioset); /* restore old tty setings */
	/* finally close */
	close((*myInterface)->IInt);
	free(*myInterface);
}


/* Set Motor state, int is 1...4, direction l,r,s for left, right, stop */
void SetMotor(IInterfaceDef **myInterface, int Motor, char direction)
{
	/*unsigned char c = 0xC1;*/
	unsigned char stop=0x3;
	char add=0; /* used for the extension Module */
	if (Motor > 4) {
		Motor-=4;
		add=1;
	}
	
	switch(direction) {
		case 'l':
			direction=0x1;
			break;
		case 'r':
			direction=0x2;
			break;
		case 's':
			direction=0x0;
			break;
		default:
			printf("Unknown Direction: %c\n", direction);
			direction=0x0;
	}
	
	/* Reset this engine to stop */
	stop<<=(Motor-1)*2; /* choose right engine and use OR to target it */
	if (add == 0) {
		(*myInterface)->MotorStatus = stop | (*myInterface)->MotorStatus;
		/* now invert 11 to 00 with a simple XOR */
		(*myInterface)->MotorStatus = stop ^ (*myInterface)->MotorStatus;
	} else {
		(*myInterface)->MotorStatus2 = stop | (*myInterface)->MotorStatus2;
		/* now invert 11 to 00 with a simple XOR */
		(*myInterface)->MotorStatus2 = stop ^ (*myInterface)->MotorStatus2;
	}
	direction<<=(Motor-1)*2; /* choose right engine */
	
	if (add == 0) {
		(*myInterface)->MotorStatus = direction | (*myInterface)->MotorStatus;
	} else {
		(*myInterface)->MotorStatus2 = direction | (*myInterface)->MotorStatus2;
	}
}


/* Returns the current engine State */
char GetMotor(IInterfaceDef **myInterface, int Motor)
{
	unsigned char state;
	char add=0; /* extension */
	
	if (Motor > 4) {
		Motor-=4;
		add=1;
	}
	
	/* Target the bytes, todo it we have to get them to the end...*/
	if (add == 0) {
		state = (*myInterface)->MotorStatus<<(4-Motor)*2;
	} else {
		state = (*myInterface)->MotorStatus2<<(4-Motor)*2;
	}
	/* and now back to the beginning */
	state>>=6;
	switch(state) {
		case 0x1:
			return 'l';
		case 0x2:
			return 'r';
		default:
			return 's';
	}
}


/* is looking at the dezimal value and converting to a value 1 oder 0 */
int GetDigital(IInterfaceDef *myInterface, int Input)
{
	unsigned char c=1;
	c<<=(Input-1);
	c&=myInterface->Digitals;
	c>>=(Input-1);
	return c;
}


/* is looking at the dezimal value and converting to a value 1 oder 0 */
int GetDigital2(IInterfaceDef *myInterface, int Input)
{
	unsigned char c=1;
	Input-=8; /* convert */
	
	c<<=(Input-1);
	c&=myInterface->Digitals2;
	c>>=(Input-1);
	return c;
}


/* Reads AnaX, Digital Input and reset the engine state */
void ReadX(IInterfaceDef **myInterface)
{
	unsigned char c = 0xC5;
	short int i;
	
	if ((*myInterface)->numIfaces == 2)
		c = 0xC6;
	
	write((*myInterface)->IInt, &c, 1);
	write((*myInterface)->IInt, &(*myInterface)->MotorStatus, 1);
	/* test for the extension module */
	
	if ((*myInterface)->numIfaces == 2)
		write((*myInterface)->IInt, &(*myInterface)->MotorStatus2, 1);
	/* ok, problem: We have to read 3 bytes, but only can get 1 per read.
	 * The first are the Digitals, the last 2 are the Analog, so read the first of the
	 * Analog, do a byte shifting and read the last */
	read((*myInterface)->IInt, &(*myInterface)->Digitals, 1);
	(*myInterface)->Digital[0] = -1;
	for (i=1; i<=8; i++)
		(*myInterface)->Digital[i] = GetDigital((*myInterface), i);
	if ((*myInterface)->numIfaces == 2) {
		read((*myInterface)->IInt, &(*myInterface)->Digitals2, 1);
		for (i=9; i<=16; i++)
			(*myInterface)->Digital[i] = GetDigital2((*myInterface), i);
	}
		
	read((*myInterface)->IInt, &(*myInterface)->AnaX, 1);
	(*myInterface)->AnaX<<=8;
	read((*myInterface)->IInt, &(*myInterface)->AnaX, 1);
}


/* same, but with Y, you can only read X _or_ Y */
void ReadY(IInterfaceDef **myInterface)
{
	unsigned char c = 0xC9;
	short int i;
	
	if ((*myInterface)->numIfaces == 2)
		c = 0xCA;
	
	write((*myInterface)->IInt, &c, 1);
	write((*myInterface)->IInt, &(*myInterface)->MotorStatus, 1);
	/* test for the extension module */
	if ((*myInterface)->numIfaces == 2)
		write((*myInterface)->IInt, &(*myInterface)->MotorStatus2, 1);
	/* ok, problem: We have to read 3 bytes, but only can get 1 per read.
	 * The first are the Digitals, the last 2 are the Analog, so read the first of the
	 * Analog, do a byte shifting and read the last */
	read((*myInterface)->IInt, &(*myInterface)->Digitals, 1);
	(*myInterface)->Digital[0] = -1;
	for (i=1; i<=8; i++)
		(*myInterface)->Digital[i] = GetDigital((*myInterface), i);
	if ((*myInterface)->numIfaces == 2) {
		read((*myInterface)->IInt, &(*myInterface)->Digitals2, 1);
		for (i=9; i<=16; i++)
			(*myInterface)->Digital[i] = GetDigital2((*myInterface), i);
	}
	read((*myInterface)->IInt, &(*myInterface)->AnaY, 1);
	(*myInterface)->AnaY<<=8;
	read((*myInterface)->IInt, &(*myInterface)->AnaY, 1);
}


/* Only read Digitals, no X, no Y - not used */
void ReadDigitals(IInterfaceDef **myInterface)
{
	unsigned char c = 0xC1;
	short int i;
	
	if ((*myInterface)->numIfaces == 2)
		c = 0xC2;
	
	write((*myInterface)->IInt, &c, 1);
	write((*myInterface)->IInt, &(*myInterface)->MotorStatus, 1);
	/* test for the extension module */
	if ((*myInterface)->numIfaces == 2)
		write((*myInterface)->IInt, &(*myInterface)->MotorStatus2, 1);
	read((*myInterface)->IInt, &(*myInterface)->Digitals, 1);
	(*myInterface)->Digital[0] = -1;
	for (i=1; i<=8; i++)
		(*myInterface)->Digital[i] = GetDigital((*myInterface), i);
	
	if ((*myInterface)->numIfaces == 2) {
		read((*myInterface)->IInt, &(*myInterface)->Digitals2, 1);
		for (i=9; i<=16; i++)
			(*myInterface)->Digital[i] = GetDigital2((*myInterface), i);
	}
}


/* Just cycle ReadX & Y to make sure the digitals, X, Y and the engine states are constantly set */
void II_thread(IInterfaceDef **myInterface)
{
	while(TRUE) {
		ReadX(myInterface);
		usleep(INTERFACE_QUERY_TIME); /* bad, but needed :( */
		ReadY(myInterface);
		usleep(INTERFACE_QUERY_TIME);
		/*ReadDigitals(myInterface);*/
	}
	
	pthread_exit((void *) 0);
}


/* Init: open the serial console and start the thread */
void IIinit(IInterfaceDef **myInterface	, char *Device, char numIfaces)
{	
	pthread_t t1;
	
	if (((*myInterface) = (IInterfaceDef *) malloc (sizeof(IInterfaceDef))) == NULL) {
		fprintf(stderr, "Error getting memory: %s\n", strerror(errno));
		exit(-1);
	}

	/* 
	* Open modem device for reading and writing and not as controlling tty
	* because we don't want to get killed if linenoise sends CTRL-C.
	*/
	/* We use Non-Canonical Input Processing!!! */
	/* O_NONBLOCK? */
	if (((*myInterface)->IInt = open(Device, O_RDWR | O_NOCTTY)) < 0) {
		/* if fail openning device, */
		fprintf(stderr, "Error openning Device:\n");
		/* push the error */
		perror(Device);
		fprintf(stderr, "exiting..\n.");
		/* and exit */
		exit(-1);
	}

	tcgetattr((*myInterface)->IInt, &(*myInterface)->saveioset); /* save current modem settings */
	
	bzero(&(*myInterface)->newioset, sizeof((*myInterface)->newioset));
	
	/* 
	 * BAUDRATE: Set bps rate. You could also use cfsetispeed and cfsetospeed.
	 * CRTSCTS : output hardware flow control (only used if the cable has all necessary lines. See sect. 7 of Serial-HOWTO)
	 * CS8     : 8n1 (8bit,no parity,1 stopbit)
	 * CLOCAL  : local connection, no modem contol
	 * CREAD   : enable receiving characters
	*/
	/* CRTSCTS? */
	(*myInterface)->newioset.c_cflag = BAUDRATE | CS8 | CLOCAL | CREAD; 

	/*
	 * IGNPAR  : ignore bytes with parity errors
	 * ICRNL   : map CR to NL (otherwise a CR input on the other computer will not terminate input)
	 * otherwise make device raw (no other input processing)
	 */
	//(*myInterface)->newioset.c_iflag = IGNPAR;
	
	/* Raw output. */
        (*myInterface)->newioset.c_oflag = 0;
        
        /* set input mode (non-canonical, no echo,...) */
        (*myInterface)->newioset.c_lflag = 0;

        (*myInterface)->newioset.c_cc[VTIME]    = 1;   /* t = TIME *0.1 s */
        (*myInterface)->newioset.c_cc[VMIN]     = 0;   /* blocking read until 1 chars received */
	
	/* now clean the modem line and activate the settings for modem */
	tcflush((*myInterface)->IInt, TCIFLUSH);
	tcsetattr((*myInterface)->IInt, TCSANOW, &(*myInterface)->newioset);
	
	(*myInterface)->numIfaces = numIfaces;
		
	if(pthread_create(&t1, NULL, (void *)&II_thread, (IInterfaceDef *)myInterface) != 0)
	{
		fprintf(stderr, "Error in Thread! - %s\n", strerror(errno));
		exit(-1);
	}
	
	/* wait 1 second before going back */
	sleep(1);
}
